import configparser
import os
import pathlib
import sys

from cargo2rpm.metadata import FeatureFlags

from rust2rpm import log
from rust2rpm.cli import get_parser
from rust2rpm.crate import process_project
from rust2rpm.cratesio import NoVersionsError
from rust2rpm.distgit import get_package_info
from rust2rpm.generator import spec_render_crate, spec_render_workspace
from rust2rpm.licensing import dump_sdpx_to_fedora_map, translate_license
from rust2rpm.utils import (
    detect_packager,
    detect_rpmautospec,
    exit_on_common_errors,
    guess_crate_name,
    package_name_suffixed,
)


@exit_on_common_errors()
def main():
    parser = get_parser()
    args = parser.parse_args()

    if args.show_license_map:
        dump_sdpx_to_fedora_map(sys.stdout)
        return

    if args.translate_license:
        license_str, comments = translate_license(args.target, args.crate)
        if comments:
            for comment in comments.split("\n"):
                log.info(comment)
        print(license_str)
        return

    if args.crate is None:
        args.crate = guess_crate_name()
        if args.crate is None:
            parser.error("crate/path argument missing and autodetection failed")

    try:
        project, diffs, metadata, doc_files, license_files = process_project(
            args.crate, args.version, args.patch, args.patch_foreign, args.store_crate
        )
    except NoVersionsError:
        log.error(f"No versions are available for crate {args.crate!r}.")
        sys.exit(1)

    if metadata.is_workspace():
        base_name = project
        rpm_name = project
    else:
        base_name = metadata.packages[0].name
        rpm_name = package_name_suffixed(base_name, args.suffix)

    patch_files = (
        f"{base_name}-fix-metadata-auto.diff" if diffs[0] else None,
        f"{base_name}-fix-metadata.diff" if diffs[1] else None,
    )

    spec_file = pathlib.Path(f"{rpm_name}.spec")

    if args.target in {"fedora"} and args.existence_check and not os.path.exists(spec_file):
        # No specfile, so this is probably a new package
        if package_info := get_package_info(rpm_name):
            if args.suffix:
                log.warn(
                    f"Version {args.suffix}.* of the crate {base_name!r} is already "
                    + f"packaged for Fedora: {package_info['full_url']}"
                )
            else:
                log.warn(f"Crate {base_name!r} is already packaged for Fedora: {package_info['full_url']}")

            log.info("Re-run with --no-existence-check to create a new spec file from scratch.")
            sys.exit(1)

    if args.rpmautospec is None:
        args.rpmautospec = detect_rpmautospec(args.target, spec_file)

    packager = detect_packager()

    conf = configparser.ConfigParser(interpolation=configparser.ExtendedInterpolation())
    confs = conf.read([".rust2rpm.conf", "_rust2rpm.conf", "rust2rpm.conf"])

    # clean up configuration files with deprecated names
    if len(confs) > 1:
        log.error(
            "More than one *rust2rpm.conf file is present in this directory. "
            + "Ensure that there is only one, and that it has the correct contents."
        )
        sys.exit(1)

    if ".rust2rpm.conf" in confs and "rust2rpm.conf" not in confs:
        os.rename(".rust2rpm.conf", "rust2rpm.conf")
        log.info("Renamed deprecated, hidden .rust2rpm.conf file to rust2rpm.conf.")

    if "_rust2rpm.conf" in confs and "rust2rpm.conf" not in confs:
        os.rename("_rust2rpm.conf", "rust2rpm.conf")
        log.info("Renamed deprecated _rust2rpm.conf file to rust2rpm.conf.")

    if args.target not in conf:
        conf.add_section(args.target)

    conf_all_features = conf[args.target].getboolean("all-features")
    if conf_all_features is False and args.all_features:
        log.warn(
            'Conflicting settings for enabling all features: The setting is "false"'
            + 'in rust2rpm.conf but it was enabled with the "--all-features" CLI flag.'
        )

    if not metadata.is_workspace():
        spec_contents = spec_render_crate(
            metadata=metadata,
            target=args.target,
            rpm_name=rpm_name,
            patch_file_automatic=patch_files[0],
            patch_file_manual=patch_files[1],
            license_files=license_files,
            doc_files=doc_files,
            distconf=conf[args.target],
            feature_flags=FeatureFlags(all_features=(conf_all_features or args.all_features)),
            relative_license_paths=args.relative_license_paths,
            rpmautospec=args.rpmautospec,
            auto_changelog_entry=args.auto_changelog_entry,
            packager=packager,
        )
    else:
        spec_contents = spec_render_workspace(
            metadata=metadata,
            target=args.target,
            rpm_name=rpm_name,
            license_files=license_files,
            doc_files=doc_files,
            distconf=conf[args.target],
            feature_flags=FeatureFlags(all_features=(conf_all_features or args.all_features)),
            rpmautospec=args.rpmautospec,
            auto_changelog_entry=args.auto_changelog_entry,
            packager=packager,
        )

    if args.stdout:
        print(f"# {spec_file}")
        print(spec_contents)
        for fname, diff in zip(patch_files, diffs):
            if fname:
                print(f"# {fname}")
                print("".join(diff), end="")
    else:
        with open(spec_file, "w") as fobj:
            fobj.write(spec_contents)
        log.success(f"Generated: {fobj.name}")
        for fname, diff in zip(patch_files, diffs):
            if fname:
                with open(fname, "w") as fobj:
                    fobj.writelines(diff)
                log.success(f"Generated: {fobj.name}")


if __name__ == "__main__":
    main()
